/**
 * Optical Sensor System for ESP32-S3
 * 
 * This program implements a high-precision optical sensing system with:
 * - Dual-wavelength LED modulation (850nm and 700nm)
 * - Dual-channel ADC sampling at 250Hz
 * - WiFi UDP data transmission
 * 
 * The system operates with a 4-phase measurement cycle:
 * Phase 0: 850nm LED ON, 700nm LED OFF
 * Phase 1: Both LEDs OFF
 * Phase 2: 850nm LED OFF, 700nm LED ON
 * Phase 3: Both LEDs OFF
 * 
 * Each phase lasts 1ms, with 8 ADC samples per phase.
 */

#include <Wire.h>
#include <WiFi.h>
#include <WiFiUdp.h>
#include <driver/adc.h>
#include <Adafruit_MCP4728.h>
#include <Preferences.h> // Added for persistent storage

// Pin Definitions
#define SDA_PIN 3
#define SCL_PIN 4

#define DEBUG_PIN_LED_A A5    // Debug pin for LED A state
#define DEBUG_PIN_LED_B A4    // Debug pin for LED B state
#define DEBUG_PIN_ADC A3      // Debug pin for ADC sampling
#define DEBUG_PIN_PROCESS A2  // Debug pin for processing
#define DEBUG_PIN_WIFI A1     // Debug pin for WiFi transmission

// LED Current Control Values
#define LED_A_VALUE 250       // 850nm LED (402mA)
#define LED_B_VALUE 90        // 700nm LED (145mA)

// Measurement Configuration
#define SAMPLES_PER_STATE 16   // Samples per LED state
#define TOGGLE_FREQUENCY 250  // Hz (250 cycles/second = 4ms per cycle)
#define PHASES_PER_CYCLE 4    // Four phases per measurement cycle
#define CLOCK_FREQ 1000000    // Clock frequency in Hz

// ADC Configuration
#define ADC0_CHANNEL ADC1_CHANNEL_8  // Pin 9
#define ADC1_CHANNEL ADC1_CHANNEL_9  // Pin 10

// Buffer Sizes
#define RESULT_BUFFER_SIZE 100
#define MEASUREMENTS_PER_PACKET 20

// Configuration for Serial timeout
#define SERIAL_WAIT_TIMEOUT 10000  // Wait up to 10 seconds for serial

// WiFi Configuration
const char* ssid = "YOUR_NETWORK";
const char* password = "YOUR_PASSWORD";
const int udpPort = 8888;

// IP address will be stored in preferences and configurable via serial
char udpAddress[16] = "0.0.0.0";  // Default placeholder
bool ipConfigured = false;

// ===== Data Structures =====

// Raw ADC data buffer
struct MeasurementData {
  // Raw ADC values [phase][sample_index][channel]
  // channel 0 = long wavelength, channel 1 = short wavelength
  uint16_t adc_buffer[PHASES_PER_CYCLE][SAMPLES_PER_STATE][2];
  
  // Current state tracking
  volatile uint8_t current_phase;        // 0-3: current phase number
  volatile uint8_t current_sample;       // 0-7: current sample within phase
  
  // Flags for synchronization
  volatile bool phase_complete;          // Set when all samples for a phase are collected
  volatile bool cycle_complete;          // Set when all phases in a cycle are complete
  
  // Timing information
  volatile unsigned long phase_start_time;   // Start time of current phase in μs
  volatile unsigned long sample_times[SAMPLES_PER_STATE]; // Precise sample timestamps
};

// Measurement result structure
struct MeasurementResult {
  int16_t channel_A_long;   // 850nm LED effect on long detector
  int16_t channel_B_long;   // 700nm LED effect on long detector
  int16_t channel_A_short;  // 850nm LED effect on short detector
  int16_t channel_B_short;  // 700nm LED effect on short detector
  uint32_t timestamp;       // Measurement timestamp in μs
};

// Result circular buffer
struct ResultBuffer {
  MeasurementResult results[RESULT_BUFFER_SIZE];
  volatile int write_pos;
  volatile int read_pos;
  volatile bool buffer_full;
};

// UDP packet structure
struct UDPPacket {
  uint32_t timestamp;      // Packet timestamp (ms)
  uint32_t period_count;   // Measurement period counter
  uint16_t packet_num;     // Current packet number
  uint16_t total_packets;  // Total packets in this group
  uint8_t reserved[4];     // Reserved for future use
  
  // Array of measurements in this packet
  struct {
    int16_t channel_A_long;
    int16_t channel_B_long;
    int16_t channel_A_short;
    int16_t channel_B_short;
    uint32_t timestamp;    // Individual measurement timestamp
  } measurements[MEASUREMENTS_PER_PACKET];
};

// ===== Global Variables =====

// Data structures
MeasurementData measurement_data;
ResultBuffer result_buffer;

// Hardware timer
hw_timer_t *timer = NULL;
portMUX_TYPE timerMux = portMUX_INITIALIZER_UNLOCKED;

// Flags for main loop
volatile bool need_sampling = false;
volatile bool need_led_change = false;

// Statistics
unsigned long total_measurements = 0;
unsigned long packets_transmitted = 0;
float measurement_rate = 0;
float adc_sampling_rate = 0;
unsigned long last_stats_time = 0;

// Data transfer statistics
unsigned long total_bytes_transmitted = 0;
unsigned long interval_bytes_transmitted = 0;
float data_transfer_rate_kb = 0.0;
unsigned long last_transfer_time = 0;

// WiFi and UDP
WiFiUDP udp;
bool wifiConnected = false;
unsigned long last_wifi_check = 0;
unsigned long last_forced_transmission = 0;

// MCP4728 DAC instance
Adafruit_MCP4728 mcp;

// Preferences for storing settings
Preferences preferences;

// Flag to signal system running
bool systemRunning = false;

// ===== Function Declarations =====
void setup_timer(int freq, int phases, int samples_per_phase);
void init_dac();
void init_buffers();
void update_led_state();
void take_adc_samples();
void process_measurement_cycle();
void store_result(MeasurementResult* result);
bool should_transmit_data();
void transmit_data();
void print_stats();
int get_buffer_usage();
void handle_serial_commands();
void setupWiFi();
void checkWiFiConnection();
void IRAM_ATTR timer_isr();
void scanI2C();
bool waitForSerial(unsigned long timeout);
bool configureIPAddress();
bool validateIP(const char* ip);
void loadSettings();
void saveSettings();

// ===== Setup Function =====
void setup() {
  // Initialize serial communication
  Serial.begin(2000000);  // 2 Mbps for high-speed debug
  
  // Wait for serial connection to be established
  bool serialConnected = waitForSerial(SERIAL_WAIT_TIMEOUT);
  
  // Set up debug pins
  pinMode(DEBUG_PIN_LED_A, OUTPUT);
  pinMode(DEBUG_PIN_LED_B, OUTPUT);
  pinMode(DEBUG_PIN_ADC, OUTPUT);
  pinMode(DEBUG_PIN_PROCESS, OUTPUT);
  pinMode(DEBUG_PIN_WIFI, OUTPUT);
  
  // Initial debug pin states (all low)
  digitalWrite(DEBUG_PIN_LED_A, LOW);
  digitalWrite(DEBUG_PIN_LED_B, LOW);
  digitalWrite(DEBUG_PIN_ADC, LOW);
  digitalWrite(DEBUG_PIN_PROCESS, LOW);
  digitalWrite(DEBUG_PIN_WIFI, LOW);
  
  Serial.println("\n=== Optical Sensor System Initializing ===");
  
  // Initialize preferences
  preferences.begin("sensor", false);
  
  // Load settings from non-volatile storage
  loadSettings();
  
  // Configure IP address if not set or if serial is connected
  if (!ipConfigured && serialConnected) {
    configureIPAddress();
  }
  
  // Set I2C power pin if your board has one
  pinMode(6, OUTPUT);
  digitalWrite(6, HIGH);
  delay(50);  // Give power time to stabilize
  
  // Set up I2C communication for MCP4728 DAC
  Wire.begin(SDA_PIN, SCL_PIN);
  Wire.setClock(400000);  // 400 kHz I2C
  
  // Initialize MCP4728 DAC
  init_dac();
  
  // Configure ADC for direct access
  adc1_config_width(ADC_WIDTH_BIT_12);
  adc1_config_channel_atten(ADC0_CHANNEL, ADC_ATTEN_DB_11);
  adc1_config_channel_atten(ADC1_CHANNEL, ADC_ATTEN_DB_11);
  
  // Initialize buffers
  init_buffers();
  
  // Connect to WiFi
  setupWiFi();
  
  // Set up timer interrupt for precise LED toggling
  setup_timer(TOGGLE_FREQUENCY, PHASES_PER_CYCLE, SAMPLES_PER_STATE);
  
  Serial.println("Initialization complete, beginning measurements");
  systemRunning = true;
}

// ===== Main Loop =====
void loop() {
  // Handle any serial commands first (prioritize user input)
  if (Serial.available()) {
    handle_serial_commands();
  }

  // Only proceed with normal operations if the system is running
  if (systemRunning) {
    // Check if we need to change LED state
    if (need_led_change) {
      // Set debug pin high during LED state changes
      digitalWrite(DEBUG_PIN_PROCESS, HIGH);
      
      update_led_state();
      need_led_change = false;
      measurement_data.phase_start_time = micros();
      
      digitalWrite(DEBUG_PIN_PROCESS, LOW);
    }
    
    // Check if we need to take ADC samples
    if (need_sampling) {
      // Set debug pin high during ADC sampling
      digitalWrite(DEBUG_PIN_ADC, HIGH);
      
      take_adc_samples();
      need_sampling = false;
      
      digitalWrite(DEBUG_PIN_ADC, LOW);
    }
    
    // Check if a full measurement cycle is complete
    if (measurement_data.cycle_complete) {
      // Set debug pin high during processing
      digitalWrite(DEBUG_PIN_PROCESS, HIGH);
      
      measurement_data.cycle_complete = false;
      process_measurement_cycle();
      
      digitalWrite(DEBUG_PIN_PROCESS, LOW);
    }
    
    // Check if we have enough data to send via UDP
    if (should_transmit_data()) {
      // Set debug pin high during transmission
      digitalWrite(DEBUG_PIN_WIFI, HIGH);
      
      transmit_data();
      
      digitalWrite(DEBUG_PIN_WIFI, LOW);
    }
    
    // Periodically check WiFi connection
    unsigned long current_time = millis();
    if (current_time - last_wifi_check > 10000) { // Check every 10 seconds
      checkWiFiConnection();
      last_wifi_check = current_time;
    }
    
    // Periodically force transmission of any pending data
    if (current_time - last_forced_transmission > 100) { // Every 100ms
      if (get_buffer_usage() > 0) {
        digitalWrite(DEBUG_PIN_WIFI, HIGH);
        transmit_data();
        digitalWrite(DEBUG_PIN_WIFI, LOW);
      }
      last_forced_transmission = current_time;
    }
    
    // Periodically update statistics
    if (current_time - last_stats_time > 5000) { // Every 5 seconds
      print_stats();
      last_stats_time = current_time;
    }
  }
}

// ===== New Functions for Serial and IP Configuration =====

// Wait for serial connection with timeout
bool waitForSerial(unsigned long timeout) {
  unsigned long startTime = millis();
  
  // Wait for serial connection or timeout
  while (!Serial && (millis() - startTime < timeout)) {
    delay(100);
  }
  
  return Serial;
}

// Configure IP address via serial
bool configureIPAddress() {
  Serial.println("\n=== IP Address Configuration ===");
  Serial.println("Please enter the UDP target IP address:");
  
  // Wait for user input with timeout
  unsigned long startTime = millis();
  while (!Serial.available() && (millis() - startTime < 30000)) {
    delay(100);
  }
  
  // Check if we timed out
  if (!Serial.available()) {
    Serial.println("Timeout waiting for IP address. Using last saved or default.");
    return false;
  }
  
  // Read the IP address from serial
  String ipString = Serial.readStringUntil('\n');
  ipString.trim();
  
  // Validate IP address
  if (validateIP(ipString.c_str())) {
    // Copy the IP to our global variable
    strncpy(udpAddress, ipString.c_str(), sizeof(udpAddress) - 1);
    udpAddress[sizeof(udpAddress) - 1] = '\0';  // Ensure null termination
    
    Serial.print("IP address set to: ");
    Serial.println(udpAddress);
    
    // Save to preferences
    saveSettings();
    
    ipConfigured = true;
    return true;
  } else {
    Serial.println("Invalid IP address format. Using last saved or default.");
    return false;
  }
}

// Validate IP address format
bool validateIP(const char* ip) {
  // Simple validation (could be improved)
  int segments = 0;
  int segmentValue = 0;
  int chars = 0;
  
  for (int i = 0; ip[i] != '\0'; i++) {
    chars++;
    
    if (ip[i] == '.') {
      segments++;
      segmentValue = 0;
    } else if (ip[i] >= '0' && ip[i] <= '9') {
      segmentValue = segmentValue * 10 + (ip[i] - '0');
      if (segmentValue > 255) {
        return false;  // Segment value too large
      }
    } else {
      return false;  // Invalid character
    }
  }
  
  // Check if we have 4 segments (3 dots) and valid length
  return (segments == 3 && chars > 6 && chars < 16);
}

// Load settings from preferences
void loadSettings() {
  // Load saved IP address if it exists
  String savedIP = preferences.getString("ip", "");
  
  if (savedIP.length() > 0 && validateIP(savedIP.c_str())) {
    strncpy(udpAddress, savedIP.c_str(), sizeof(udpAddress) - 1);
    udpAddress[sizeof(udpAddress) - 1] = '\0';  // Ensure null termination
    ipConfigured = true;
    
    Serial.print("Loaded saved IP address: ");
    Serial.println(udpAddress);
  } else {
    Serial.println("No valid IP address found in storage.");
    ipConfigured = false;
  }
}

// Save settings to preferences
void saveSettings() {
  preferences.putString("ip", udpAddress);
  Serial.println("Settings saved.");
}

// ===== Timer and Interrupt Functions =====

// Timer setup function
void setup_timer(int freq, int phases, int samples_per_phase) {
  // Calculate period in microseconds
  // For 250Hz with 4 phases and 8 samples per phase:
  // Period = 1,000,000 / (250 * 4 * 8) = 125μs per sample
  uint64_t periodMicros = 1000000 / (freq * phases * samples_per_phase);
  
  // Initialize timer with prescaler
  timer = timerBegin(CLOCK_FREQ);
  timerAttachInterrupt(timer, &timer_isr);
  timerAlarm(timer, periodMicros, true, 0);
  
  Serial.printf("Timer set with period of %llu μs\n", periodMicros);
  Serial.printf("Expected sampling rate: %d Hz\n", freq);
}

// Timer ISR function - keep this as short as possible
void IRAM_ATTR timer_isr() {
  portENTER_CRITICAL_ISR(&timerMux);
  
  // Set flag to signal sampling needed
  need_sampling = true;
  
  // Increment sample counter
  if (++measurement_data.current_sample >= SAMPLES_PER_STATE) {
    // Move to next phase when all samples are collected
    measurement_data.current_sample = 0;
    measurement_data.phase_complete = true;
    
    // Increment phase or mark cycle complete
    if (++measurement_data.current_phase >= PHASES_PER_CYCLE) {
      measurement_data.current_phase = 0;
      measurement_data.cycle_complete = true;
    }
    
    // Signal LED state change needed
    need_led_change = true;
  }
  
  portEXIT_CRITICAL_ISR(&timerMux);
}

// ===== Initialization Functions =====

// Initialize DAC
void init_dac() {
  // Initialize MCP4728 with retry
  int retryCount = 0;
  bool initialized = false;
  
  while (!initialized && retryCount < 5) {
    if (mcp.begin()) {
      initialized = true;
      Serial.println("MCP4728 initialized successfully!");
    } else {
      retryCount++;
      Serial.print("Failed to initialize MCP4728. Retry ");
      Serial.print(retryCount);
      Serial.println("/5");
      delay(100);
    }
  }
  
  if (!initialized) {
    Serial.println("Could not initialize MCP4728. Scanning I2C bus...");
    scanI2C();
    Serial.println("Continuing anyway...");
  }
  
  // Reset all channels to 0
  mcp.setChannelValue(MCP4728_CHANNEL_A, 0);
  mcp.setChannelValue(MCP4728_CHANNEL_B, 0);
  mcp.setChannelValue(MCP4728_CHANNEL_C, 0);
  mcp.setChannelValue(MCP4728_CHANNEL_D, 0);
  
  Serial.println("DAC initialized");
}

// Initialize buffers
void init_buffers() {
  // Clear measurement data
  memset(&measurement_data, 0, sizeof(measurement_data));
  
  // Initialize result buffer
  result_buffer.write_pos = 0;
  result_buffer.read_pos = 0;
  result_buffer.buffer_full = false;
  
  Serial.println("Buffers initialized");
}

// Set up WiFi connection
void setupWiFi() {
  // Only attempt connection if we have a valid IP configured
  if (!ipConfigured) {
    Serial.println("Cannot establish WiFi connection: IP address not configured");
    wifiConnected = false;
    return;
  }
  
  Serial.printf("Connecting to WiFi network: %s\n", ssid);
  
  WiFi.begin(ssid, password);
  
  // Try for 10 seconds
  int retries = 20; // 20 x 500ms = 10 seconds
  while (WiFi.status() != WL_CONNECTED && retries > 0) {
    delay(500);
    Serial.print(".");
    retries--;
  }
  
  if (WiFi.status() == WL_CONNECTED) {
    Serial.println("\nWiFi connected");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
    wifiConnected = true;
    
    // Start UDP
    udp.begin(udpPort);
    Serial.printf("UDP initialized on port %d\n", udpPort);
    Serial.printf("UDP target set to: %s\n", udpAddress);
  } else {
    Serial.println("\nWiFi connection failed");
    wifiConnected = false;
  }
}

// Check WiFi connection and reconnect if needed
void checkWiFiConnection() {
  if (!ipConfigured) {
    return;  // Don't try to connect if IP isn't configured
  }
  
  if (WiFi.status() != WL_CONNECTED) {
    if (wifiConnected) {
      Serial.println("WiFi connection lost, attempting to reconnect");
      wifiConnected = false;
    }
    
    WiFi.begin(ssid, password);
    
    // Check if connected after a short delay
    delay(500);
    
    if (WiFi.status() == WL_CONNECTED) {
      Serial.println("WiFi reconnected");
      Serial.print("IP address: ");
      Serial.println(WiFi.localIP());
      wifiConnected = true;
      
      // Restart UDP
      udp.begin(udpPort);
    }
  }
}

// ===== LED Control and ADC Sampling =====

// LED state control function
void update_led_state() {
  switch (measurement_data.current_phase) {
    case 0:  // Phase 0: LED A ON, LED B OFF
      // Update DAC values
      mcp.setChannelValue(MCP4728_CHANNEL_A, LED_A_VALUE);
      mcp.setChannelValue(MCP4728_CHANNEL_B, 0);
      
      // Update debug pins
      digitalWrite(DEBUG_PIN_LED_A, HIGH);
      digitalWrite(DEBUG_PIN_LED_B, LOW);
      break;
      
    case 1:  // Phase 1: LED A OFF, LED B OFF
      mcp.setChannelValue(MCP4728_CHANNEL_A, 0);
      
      // Update debug pins
      digitalWrite(DEBUG_PIN_LED_A, LOW);
      digitalWrite(DEBUG_PIN_LED_B, LOW);
      break;
      
    case 2:  // Phase 2: LED A OFF, LED B ON
      mcp.setChannelValue(MCP4728_CHANNEL_B, LED_B_VALUE);
      
      // Update debug pins
      digitalWrite(DEBUG_PIN_LED_A, LOW);
      digitalWrite(DEBUG_PIN_LED_B, HIGH);
      break;
      
    case 3:  // Phase 3: LED A OFF, LED B OFF
      mcp.setChannelValue(MCP4728_CHANNEL_B, 0);
      
      // Update debug pins
      digitalWrite(DEBUG_PIN_LED_A, LOW);
      digitalWrite(DEBUG_PIN_LED_B, LOW);
      break;
  }
}

// ADC sampling function
void take_adc_samples() {
  // Record sample timestamp
  measurement_data.sample_times[measurement_data.current_sample] = micros() - measurement_data.phase_start_time;
  
  // Sample both channels in sequence (as fast as possible)
  uint16_t long_sample = adc1_get_raw(ADC0_CHANNEL);
  uint16_t short_sample = adc1_get_raw(ADC1_CHANNEL);
  
  // Store samples in buffer
  measurement_data.adc_buffer[measurement_data.current_phase][measurement_data.current_sample][0] = long_sample;
  measurement_data.adc_buffer[measurement_data.current_phase][measurement_data.current_sample][1] = short_sample;
}

// ===== Data Processing =====

// Process data from a complete measurement cycle
void process_measurement_cycle() {
  // Calculate averages for each phase and channel
  uint16_t avg_long[PHASES_PER_CYCLE];
  uint16_t avg_short[PHASES_PER_CYCLE];
  
  for (int phase = 0; phase < PHASES_PER_CYCLE; phase++) {
    // Fast bit-shift averaging for long wavelength
    uint32_t sum_long = 0;
    uint32_t sum_short = 0;
    
    // Unrolled loop for better performance
    for (int i = 0; i < SAMPLES_PER_STATE; i++) {
      sum_long += measurement_data.adc_buffer[phase][i][0];
      sum_short += measurement_data.adc_buffer[phase][i][1];
    }
    
    // Divide by 16 using bit-shift (much faster than division)
    avg_long[phase] = sum_long >> 4;
    avg_short[phase] = sum_short >> 4;
  }
  
  // Create result structure
  MeasurementResult result;
  
  // Calculate differences (on - off)
  result.channel_A_long = avg_long[0] - avg_long[1];    // Phase 0 - Phase 1 (850nm effect on long detector)
  result.channel_B_long = avg_long[2] - avg_long[3];    // Phase 2 - Phase 3 (700nm effect on long detector)
  
  result.channel_A_short = avg_short[0] - avg_short[1]; // Phase 0 - Phase 1 (850nm effect on short detector)
  result.channel_B_short = avg_short[2] - avg_short[3]; // Phase 2 - Phase 3 (700nm effect on short detector)
  
  // Add timestamp
  result.timestamp = micros();
  
  // Only store the result if all data points are greater than zero
  if (result.channel_A_short > 0 && 
      result.channel_B_short > 0) {
    // Store result in buffer
    store_result(&result);
    
    // Update total measurement count
    total_measurements++;
  } else {
    // Optionally, add debug output for discarded measurements
    // Serial.println("Discarded measurement with non-positive values");
  }
}

// Store measurement result in circular buffer
void store_result(MeasurementResult* result) {
  // Copy result to buffer
  result_buffer.results[result_buffer.write_pos] = *result;
  
  // Update write position
  result_buffer.write_pos = (result_buffer.write_pos + 1) % RESULT_BUFFER_SIZE;
  
  // Check if buffer is full
  if (result_buffer.write_pos == result_buffer.read_pos) {
    result_buffer.buffer_full = true;
  }
}

// ===== Data Transmission =====

// Check if we should transmit data
bool should_transmit_data() {
  // Don't transmit if IP isn't configured
  if (!ipConfigured) {
    return false;
  }
  
  // Transmit if buffer is full
  if (result_buffer.buffer_full) {
    return true;
  }
  
  // Calculate available data (accounting for wrap-around)
  int available = get_buffer_usage();
  
  // Transmit if we have at least one packet worth of data or 75% buffer capacity
  return (available >= MEASUREMENTS_PER_PACKET) || 
         (available >= (RESULT_BUFFER_SIZE * 3 / 4));
}

// Transmit data via UDP
void transmit_data() {
  // Only proceed if WiFi is connected and IP is configured
  if (!wifiConnected || !ipConfigured) {
    if (!ipConfigured) {
      Serial.println("Cannot transmit: IP not configured");
    } else {
      Serial.println("Cannot transmit: WiFi not connected");
    }
    return;
  }
  
  // Calculate how many complete packets we can send
  int available = get_buffer_usage();
  
  int packets_to_send = available / MEASUREMENTS_PER_PACKET;
  if (packets_to_send == 0 && available > 0) {
    packets_to_send = 1;  // Send at least one packet if there's data
  }
  
  if (packets_to_send == 0) {
    return;  // Nothing to send
  }
  
  // Prepare and send packets
  for (int packet_num = 0; packet_num < packets_to_send; packet_num++) {
    UDPPacket packet;
    
    // Set packet header
    packet.timestamp = millis();
    packet.period_count = total_measurements;
    packet.packet_num = packet_num;
    packet.total_packets = packets_to_send;
    memset(packet.reserved, 0, sizeof(packet.reserved));
    
    // Calculate how many measurements to include in this packet
    int measurements_in_packet = MEASUREMENTS_PER_PACKET;
    if (packet_num == packets_to_send - 1 && available % MEASUREMENTS_PER_PACKET != 0) {
      measurements_in_packet = available % MEASUREMENTS_PER_PACKET;
    }
    
    // Fill packet with measurements
    for (int i = 0; i < measurements_in_packet; i++) {
      int idx = (result_buffer.read_pos + i) % RESULT_BUFFER_SIZE;
      
      packet.measurements[i].channel_A_long = result_buffer.results[idx].channel_A_long;
      packet.measurements[i].channel_B_long = result_buffer.results[idx].channel_B_long;
      packet.measurements[i].channel_A_short = result_buffer.results[idx].channel_A_short;
      packet.measurements[i].channel_B_short = result_buffer.results[idx].channel_B_short;
      packet.measurements[i].timestamp = result_buffer.results[idx].timestamp;
    }
    
    // Calculate packet size in bytes
    int packet_size = 16 + (measurements_in_packet * 12); // Header + measurements
    
    // Send packet
    udp.beginPacket(udpAddress, udpPort);
    udp.write((uint8_t*)&packet, packet_size);
    udp.endPacket();
    
    // Update read position
    result_buffer.read_pos = (result_buffer.read_pos + measurements_in_packet) % RESULT_BUFFER_SIZE;
    
    // Update statistics
    packets_transmitted++;
    total_bytes_transmitted += packet_size;
    interval_bytes_transmitted += packet_size;
    last_transfer_time = millis();
  }
  
  // Reset buffer full flag
  result_buffer.buffer_full = false;
}

// ===== Utilities and Monitoring =====

// Print debug statistics
void print_stats() {
  unsigned long current_time = millis();
  
  // Calculate measurement rate (over last 5 seconds)
  static unsigned long last_total_measurements = 0;
  measurement_rate = (total_measurements - last_total_measurements) / 5.0;
  last_total_measurements = total_measurements;
  
  // Calculate ADC sampling rate
  adc_sampling_rate = measurement_rate * PHASES_PER_CYCLE * SAMPLES_PER_STATE * 2; // 2 channels
  
  // Calculate data transfer rate (KB/sec)
  float elapsed_sec = (current_time - last_stats_time) / 1000.0;
  if (elapsed_sec > 0) {
    data_transfer_rate_kb = (interval_bytes_transmitted / 1024.0) / elapsed_sec;
  }
  interval_bytes_transmitted = 0; // Reset for next interval
  
  Serial.println("=== System Statistics ===");
  Serial.printf("Measurement rate: %.2f Hz\n", measurement_rate);
  Serial.printf("ADC sampling rate: %.2f kHz\n", adc_sampling_rate / 1000.0);
  Serial.printf("Data transfer rate: %.2f KB/sec\n", data_transfer_rate_kb);
  Serial.printf("Buffer usage: %d/%d (%.1f%%)\n", 
    get_buffer_usage(), 
    RESULT_BUFFER_SIZE, 
    (float)get_buffer_usage() / RESULT_BUFFER_SIZE * 100);
  Serial.printf("Total measurements: %lu\n", total_measurements);
  Serial.printf("Packets transmitted: %lu\n", packets_transmitted);
  Serial.printf("Total data transmitted: %.2f KB\n", total_bytes_transmitted / 1024.0);
  Serial.printf("Current phase: %d\n", measurement_data.current_phase);
  Serial.printf("WiFi status: %s\n", WiFi.status() == WL_CONNECTED ? "Connected" : "Disconnected");
  Serial.printf("UDP target IP: %s\n", ipConfigured ? udpAddress : "Not configured");
  
  // Print latest measurement
  if (result_buffer.write_pos != result_buffer.read_pos || result_buffer.buffer_full) {
    int latest_idx = (result_buffer.write_pos > 0) ? 
                     (result_buffer.write_pos - 1) : 
                     (RESULT_BUFFER_SIZE - 1);
    
    MeasurementResult* latest = &result_buffer.results[latest_idx];
    
    Serial.println("Latest Measurement:");
    Serial.printf("  A_long: %d\n", latest->channel_A_long);
    Serial.printf("  B_long: %d\n", latest->channel_B_long);
    Serial.printf("  A_short: %d\n", latest->channel_A_short);
    Serial.printf("  B_short: %d\n", latest->channel_B_short);
    Serial.printf("  Timestamp: %lu\n", latest->timestamp);
  }
}

// Calculate buffer usage
int get_buffer_usage() {
  if (result_buffer.buffer_full) {
    return RESULT_BUFFER_SIZE;
  }
  
  if (result_buffer.write_pos >= result_buffer.read_pos) {
    return result_buffer.write_pos - result_buffer.read_pos;
  } else {
    return RESULT_BUFFER_SIZE - (result_buffer.read_pos - result_buffer.write_pos);
  }
}

// Handle serial commands
void handle_serial_commands() {
  String command = Serial.readStringUntil('\n');
  command.trim();
  
  if (command == "stats") {
    print_stats();
  }
  else if (command == "adc") {
    // Test ADC channels directly
    Serial.println("Testing ADC channels...");
    
    Serial.println("Reading ADC channels directly:");
    int val0 = adc1_get_raw(ADC0_CHANNEL);
    int val1 = adc1_get_raw(ADC1_CHANNEL);
    
    Serial.print("ADC0 (");
    Serial.print(ADC0_CHANNEL);
    Serial.print("): ");
    Serial.println(val0);
    
    Serial.print("ADC1 (");
    Serial.print(ADC1_CHANNEL);
    Serial.print("): ");
    Serial.println(val1);
  }
  else if (command == "led") {
    // Test LED control
    Serial.println("Testing LEDs...");
    
    // Turn on LED A
    Serial.println("LED A ON, LED B OFF");
    mcp.setChannelValue(MCP4728_CHANNEL_A, LED_A_VALUE);
    mcp.setChannelValue(MCP4728_CHANNEL_B, 0);
    
    delay(1000);
    
    // Turn on LED B
    Serial.println("LED A OFF, LED B ON");
    mcp.setChannelValue(MCP4728_CHANNEL_A, 0);
    mcp.setChannelValue(MCP4728_CHANNEL_B, LED_B_VALUE);
    
    delay(1000);
    
    // Turn off all LEDs
    Serial.println("All LEDs OFF");
    mcp.setChannelValue(MCP4728_CHANNEL_A, 0);
    mcp.setChannelValue(MCP4728_CHANNEL_B, 0);
  }
  else if (command == "timing") {
    // Print detailed timing information
    Serial.println("Timing Information:");
    Serial.printf("Current phase: %d\n", measurement_data.current_phase);
    Serial.printf("Current sample: %d\n", measurement_data.current_sample);
    Serial.println("Sample times within current phase (μs):");
    for (int i = 0; i < SAMPLES_PER_STATE; i++) {
      Serial.printf("  Sample %d: %lu\n", i, measurement_data.sample_times[i]);
    }
  }
  else if (command == "buffer") {
    // Display current buffer contents
    Serial.printf("Buffer usage: %d/%d\n", get_buffer_usage(), RESULT_BUFFER_SIZE);
    Serial.printf("Write position: %d\n", result_buffer.write_pos);
    Serial.printf("Read position: %d\n", result_buffer.read_pos);
    Serial.printf("Buffer full flag: %s\n", result_buffer.buffer_full ? "true" : "false");
    
    int count = min(10, get_buffer_usage());
    if (count > 0) {
      Serial.printf("Last %d measurements:\n", count);
      for (int i = 1; i <= count; i++) {
        int idx = (result_buffer.write_pos - i + RESULT_BUFFER_SIZE) % RESULT_BUFFER_SIZE;
        MeasurementResult* result = &result_buffer.results[idx];
        Serial.printf("  [%d] A_long: %d, B_long: %d, A_short: %d, B_short: %d, Time: %lu\n",
          i, result->channel_A_long, result->channel_B_long, 
          result->channel_A_short, result->channel_B_short,
          result->timestamp);
      }
    }
  }
  else if (command == "send") {
    // Force a test packet transmission
    Serial.println("Forcing packet transmission...");
    result_buffer.buffer_full = true;
    transmit_data();
  }
  else if (command == "ip") {
    // Configure IP address
    Serial.println("Current IP address: " + String(udpAddress));
    Serial.println("Please enter new IP address:");
    
    // Wait for input
    while (!Serial.available()) {
      delay(100);
    }
    
    String newIP = Serial.readStringUntil('\n');
    newIP.trim();
    
    if (validateIP(newIP.c_str())) {
      strncpy(udpAddress, newIP.c_str(), sizeof(udpAddress) - 1);
      udpAddress[sizeof(udpAddress) - 1] = '\0';  // Ensure null termination
      
      Serial.println("IP address updated to: " + String(udpAddress));
      ipConfigured = true;
      
      // Save to preferences
      saveSettings();
      
      // Restart WiFi with new IP
      setupWiFi();
    } else {
      Serial.println("Invalid IP address format. IP not changed.");
    }
  }
  else if (command == "restart") {
    // Restart the device
    Serial.println("Restarting device...");
    delay(1000);
    ESP.restart();
  }
  else if (command == "pause") {
    // Pause measurements
    systemRunning = false;
    Serial.println("System paused. Type 'resume' to continue.");
  }
  else if (command == "resume") {
    // Resume measurements
    systemRunning = true;
    Serial.println("System resumed.");
  }
  else {
    Serial.println("Unknown command. Available commands:");
    Serial.println("  stats - Print system statistics");
    Serial.println("  adc - Test ADC readings directly");
    Serial.println("  led - Test LED control");
    Serial.println("  timing - Print detailed timing information");
    Serial.println("  buffer - Display current buffer contents");
    Serial.println("  send - Force a test packet transmission");
    Serial.println("  ip - Configure UDP target IP address");
    Serial.println("  restart - Restart the device");
    Serial.println("  pause - Pause measurements");
    Serial.println("  resume - Resume measurements");
  }
}

// Scan the I2C bus for devices
void scanI2C() {
  Serial.println("Scanning I2C bus...");
  byte error, address;
  int deviceCount = 0;
  
  for (address = 1; address < 127; address++) {
    Wire.beginTransmission(address);
    error = Wire.endTransmission();
    
    if (error == 0) {
      Serial.print("I2C device found at address 0x");
      if (address < 16) Serial.print("0");
      Serial.print(address, HEX);
      Serial.print(" (");
      Serial.print(address);
      Serial.println(")");
      deviceCount++;
    }
  }
  
  if (deviceCount == 0) {
    Serial.println("No I2C devices found!");
  } else {
    Serial.print("Found ");
    Serial.print(deviceCount);
    Serial.println(" device(s)");
  }
}